// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

mod input_stream;
pub use self::input_stream::InputStream;

mod output_stream;
pub use self::output_stream::OutputStream;

pub(crate) mod functions;

pub(crate) mod traits {
    pub use super::input_stream::Win32InputStreamExt;
    pub use super::output_stream::Win32OutputStreamExt;
}
