// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

#include "manual.h"
#include <stdalign.h>
#include <stdio.h>

int main() {
    printf("%s;%zu;%zu\n", "GWin32InputStream", sizeof(GWin32InputStream), alignof(GWin32InputStream));
    printf("%s;%zu;%zu\n", "GWin32InputStreamClass", sizeof(GWin32InputStreamClass), alignof(GWin32InputStreamClass));
    printf("%s;%zu;%zu\n", "GWin32NetworkMonitor", sizeof(GWin32NetworkMonitor), alignof(GWin32NetworkMonitor));
    printf("%s;%zu;%zu\n", "GWin32NetworkMonitorClass", sizeof(GWin32NetworkMonitorClass), alignof(GWin32NetworkMonitorClass));
    printf("%s;%zu;%zu\n", "GWin32OutputStream", sizeof(GWin32OutputStream), alignof(GWin32OutputStream));
    printf("%s;%zu;%zu\n", "GWin32OutputStreamClass", sizeof(GWin32OutputStreamClass), alignof(GWin32OutputStreamClass));
    return 0;
}
